'use strict';

var Ajv = require('ajv');
var path = require('path');
var ts = require('typescript');
var parser_js = require('react-docgen-typescript/lib/parser.js');

/**
 * Creates an error message using the provided error object or error message. It
 * wraps the message with the name of the plugin to make it easier to track down
 * issues when debugging build errors.
 *
 * @param error Error message or error object to wrap.
 */
function createAppError(error) {
    var errorMessage = error
        ? typeof error === "string" ? error : error.message
        : "An unexpected error has occurred.";
    var appError = new Error("react-docgen-typescript-loader: " + errorMessage);
    appError.innerError = typeof error === "string" ? undefined : error;
    Error.captureStackTrace(appError, createAppError);
    return appError;
}

var ajv = new Ajv();
var validate = ajv.compile({
    type: "object",
    additionalProperties: false,
    properties: {
        skipPropsWithName: {
            anyOf: [
                { type: "string", minLength: 1 },
                { type: "array", minItems: 1, items: { type: "string", minLength: 1 } },
            ],
        },
        skipPropsWithoutDoc: {
            type: "boolean",
        },
        tsconfigPath: {
            type: "string",
            minLength: 1,
        },
        compilerOptions: {
            type: "object",
        },
        docgenCollectionName: {
            anyOf: [{ type: "string", minLength: 1 }, { type: "null" }],
        },
    },
});
function validateOptions(options) {
    if (!options)
        return;
    var valid = validate(options);
    if (!valid) {
        var errorMessage = validate.errors
            .map(function (e) {
            var message = e.message;
            if (e.dataPath)
                message = e.dataPath + " " + message;
            return message;
        })
            .join(", ");
        throw createAppError(errorMessage);
    }
}

function generateDocgenCodeBlock (filename, source, componentDocs, docgenCollectionName) {
    var sourceFile = ts.createSourceFile(filename, source, ts.ScriptTarget.ESNext);
    var relativeFilename = path.relative("./", path.resolve("./", filename))
        .replace(/\\/g, "/");
    var wrapInTryStatement = function (statements) {
        return ts.createTry(ts.createBlock(statements, true), ts.createCatchClause(ts.createVariableDeclaration(ts.createIdentifier("__react_docgen_typescript_loader_error")), ts.createBlock([])), undefined);
    };
    var codeBlocks = componentDocs.map(function (d) {
        return wrapInTryStatement([
            setDisplayName(d),
            setComponentDocGen(d),
            docgenCollectionName != null
                ? insertDocgenIntoGlobalCollection(d, docgenCollectionName, relativeFilename)
                : null,
        ].filter(function (s) { return s !== null; }));
    });
    var printer = ts.createPrinter({ newLine: ts.NewLineKind.LineFeed });
    var printNode = function (sourceNode) {
        return printer.printNode(ts.EmitHint.Unspecified, sourceNode, sourceFile);
    };
    // Concat original source code with code from generated code blocks.
    var result = codeBlocks.reduce(function (acc, node) { return acc + printNode(node); }, printNode(sourceFile));
    return result;
}
/**
 * Set component display name.
 *
 * ```
 * SimpleComponent.displayName = "SimpleComponent";
 * ```
 */
function setDisplayName(d) {
    return insertTsIgnoreBeforeStatement(ts.createStatement(ts.createBinary(ts.createPropertyAccess(ts.createIdentifier(d.displayName), ts.createIdentifier("displayName")), ts.SyntaxKind.EqualsToken, ts.createLiteral(d.displayName))));
}
/**
 * Sets the field `__docgenInfo` for the component specified by the component
 * doc with the docgen information.
 *
 * ```
 * SimpleComponent.__docgenInfo = {
 *   description: ...,
 *   displayName: ...,
 *   props: ...,
 * }
 * ```
 *
 * @param d Component doc.
 */
function setComponentDocGen(d) {
    return insertTsIgnoreBeforeStatement(ts.createStatement(ts.createBinary(
    // SimpleComponent.__docgenInfo
    ts.createPropertyAccess(ts.createIdentifier(d.displayName), ts.createIdentifier("__docgenInfo")), ts.SyntaxKind.EqualsToken, ts.createObjectLiteral([
        // SimpleComponent.__docgenInfo.description
        ts.createPropertyAssignment(ts.createLiteral("description"), ts.createLiteral(d.description)),
        // SimpleComponent.__docgenInfo.displayName
        ts.createPropertyAssignment(ts.createLiteral("displayName"), ts.createLiteral(d.displayName)),
        // SimpleComponent.__docgenInfo.props
        ts.createPropertyAssignment(ts.createLiteral("props"), ts.createObjectLiteral(Object.entries(d.props).map(function (_a) {
            var propName = _a[0], prop = _a[1];
            return createPropDefinition(propName, prop);
        }))),
    ]))));
}
/**
 * Set a component prop description.
 * ```
 * SimpleComponent.__docgenInfo.props.someProp = {
 *   defaultValue: "blue",
 *   description: "Prop description.",
 *   name: "someProp",
 *   required: true,
 *   type: "'blue' | 'green'",
 * }
 * ```
 *
 * @param propName Prop name
 * @param prop Prop definition from `ComponentDoc.props`
 */
function createPropDefinition(propName, prop) {
    /**
     * Set default prop value.
     *
     * ```
     * SimpleComponent.__docgenInfo.props.someProp.defaultValue = null;
     * SimpleComponent.__docgenInfo.props.someProp.defaultValue = {
     *   value: "blue",
     * };
     * ```
     *
     * @param defaultValue Default prop value or null if not set.
     */
    var setDefaultValue = function (defaultValue) {
        return ts.createPropertyAssignment(ts.createLiteral("defaultValue"), 
        // Use a more extensive check on defaultValue. Sometimes the parser
        // returns an empty object.
        defaultValue != null &&
            typeof defaultValue === "object" &&
            "value" in defaultValue &&
            typeof defaultValue.value === "string"
            ? ts.createObjectLiteral([
                ts.createPropertyAssignment(ts.createIdentifier("value"), ts.createLiteral(defaultValue.value)),
            ])
            : ts.createNull());
    };
    /** Set a property with a string value */
    var setStringLiteralField = function (fieldName, fieldValue) {
        return ts.createPropertyAssignment(ts.createLiteral(fieldName), ts.createLiteral(fieldValue));
    };
    /**
     * ```
     * SimpleComponent.__docgenInfo.props.someProp.description = "Prop description.";
     * ```
     * @param description Prop description.
     */
    var setDescription = function (description) {
        return setStringLiteralField("description", description);
    };
    /**
     * ```
     * SimpleComponent.__docgenInfo.props.someProp.name = "someProp";
     * ```
     * @param name Prop name.
     */
    var setName = function (name) { return setStringLiteralField("name", name); };
    /**
     * ```
     * SimpleComponent.__docgenInfo.props.someProp.required = true;
     * ```
     * @param required Whether prop is required or not.
     */
    var setRequired = function (required) {
        return ts.createPropertyAssignment(ts.createLiteral("required"), required ? ts.createTrue() : ts.createFalse());
    };
    /**
     * ```
     * SimpleComponent.__docgenInfo.props.someProp.type = { name: "'blue' | 'green'"}
     * ```
     * @param typeName Prop type name.
     */
    var setType = function (typeName) {
        return ts.createPropertyAssignment(ts.createLiteral("type"), ts.createObjectLiteral([setStringLiteralField("name", typeName)]));
    };
    return ts.createPropertyAssignment(ts.createLiteral(propName), ts.createObjectLiteral([
        setDefaultValue(prop.defaultValue),
        setDescription(prop.description),
        setName(prop.name),
        setRequired(prop.required),
        setType(prop.type.name),
    ]));
}
/**
 * Adds a component's docgen info to the global docgen collection.
 *
 * ```
 * if (typeof STORYBOOK_REACT_CLASSES !== "undefined") {
 *   STORYBOOK_REACT_CLASSES["src/.../SimpleComponent.tsx"] = {
 *     name: "SimpleComponent",
 *     docgenInfo: SimpleComponent.__docgenInfo,
 *     path: "src/.../SimpleComponent.tsx",
 *   };
 * }
 * ```
 *
 * @param d Component doc.
 * @param docgenCollectionName Global docgen collection variable name.
 * @param relativeFilename Relative file path of the component source file.
 */
function insertDocgenIntoGlobalCollection(d, docgenCollectionName, relativeFilename) {
    return insertTsIgnoreBeforeStatement(ts.createIf(ts.createBinary(ts.createTypeOf(ts.createIdentifier(docgenCollectionName)), ts.SyntaxKind.ExclamationEqualsEqualsToken, ts.createLiteral("undefined")), insertTsIgnoreBeforeStatement(ts.createStatement(ts.createBinary(ts.createElementAccess(ts.createIdentifier(docgenCollectionName), ts.createLiteral(relativeFilename + "#" + d.displayName)), ts.SyntaxKind.EqualsToken, ts.createObjectLiteral([
        ts.createPropertyAssignment(ts.createIdentifier("docgenInfo"), ts.createPropertyAccess(ts.createIdentifier(d.displayName), ts.createIdentifier("__docgenInfo"))),
        ts.createPropertyAssignment(ts.createIdentifier("name"), ts.createLiteral(d.displayName)),
        ts.createPropertyAssignment(ts.createIdentifier("path"), ts.createLiteral(relativeFilename + "#" + d.displayName)),
    ]))))));
}
/**
 * Inserts a ts-ignore comment above the supplied statement.
 *
 * It is used to work around type errors related to fields like __docgenInfo not
 * being defined on types. It also prevents compile errors related to attempting
 * to assign to nonexistent components, which can happen due to incorrect
 * detection of component names when using the parser.
 * ```
 * // @ts-ignore
 * ```
 * @param statement
 */
function insertTsIgnoreBeforeStatement(statement) {
    ts.setSyntheticLeadingComments(statement, [
        {
            text: " @ts-ignore",
            kind: ts.SyntaxKind.SingleLineCommentTrivia,
            pos: -1,
            end: -1,
        },
    ]);
    return statement;
}

var loader = function (source) {
    var _this = this;
    var callback = this.async();
    try {
        // Mark the loader as being cacheable since the result should be
        // deterministic.
        this.cacheable(true);
        var options_1 = this.query || {};
        validateOptions(options_1);
        options_1.docgenCollectionName =
            options_1.docgenCollectionName || "STORYBOOK_REACT_CLASSES";
        // Convert loader's flat options into expected structure for
        // react-docgen-typescript.
        // See: node_modules/react-docgen-typescript/lib/parser.d.ts
        var parserOptions_1 = {
            propFilter: options_1.skipPropsWithName || options_1.skipPropsWithoutDoc
                ? {
                    skipPropsWithName: options_1.skipPropsWithName || undefined,
                    skipPropsWithoutDoc: options_1.skipPropsWithoutDoc || undefined,
                }
                : undefined,
        };
        var newSource = (function () {
            // Configure parser using settings provided to loader.
            // See: node_modules/react-docgen-typescript/lib/parser.d.ts
            var parser = parser_js.withDefaultConfig(parserOptions_1);
            if (options_1.tsconfigPath)
                parser = parser_js.withCustomConfig(options_1.tsconfigPath, parserOptions_1);
            else if (options_1.compilerOptions)
                parser = parser_js.withCompilerOptions(options_1.compilerOptions, parserOptions_1);
            var componentDocs = parser.parse(_this.resourcePath);
            // Return amended source code if there is docgen information available.
            if (componentDocs.length) {
                return generateDocgenCodeBlock(_this.resourcePath, source, componentDocs, options_1.docgenCollectionName);
            }
            // Return unchanged source code if no docgen information was available.
            return source;
        })();
        if (!callback)
            return newSource;
        callback(null, newSource);
        return;
    }
    catch (e) {
        var wrappedError = createAppError(e);
        if (callback) {
            callback(wrappedError);
            return;
        }
        throw wrappedError;
    }
};

module.exports = loader;
